@@ -8,28 +8,51 @@ rules_cert_id = [
     "BSI-DSZ-CC-[0-9]+?-[0-9]+",  # German BSI
     "BSI-DSZ-CC-[0-9]+?-(?:V|v)[0-9]+-[0-9]+",  # German BSI
     "BSI-DSZ-CC-[0-9]+?-(?:V|v)[0-9]+",  # German BSI
+    "BSI-DSZ-CC-[0-9]+-(?:V|v)[0-9]+(-[0-9][0-9][0-9][0-9])*",  # German BSI (number + version + year or without year)
+    "BSI-DSZ-CC-[0-9]+-[0-9][0-9][0-9][0-9]",  # German BSI (number + year, no version)
+    "BSI-DSZ-CC-[0-9]+-(?:V|v)[0-9]+(?!-)",  # German BSI (number + version but no year => no - after version)
     # 'CC-Zert-.+?',
-    "ANSSI(?:-|-CC-)[0-9]+?/[0-9]+",  # French
+    "ANSSI(?:-|-CC-)[0-9]+?/[0-9]+(v[1-9])?",  # French
+    "ANSS[Ii]-CC-[0-9][0-9][0-9][0-9][/-_][0-9][0-9]+(?!-M|-S)",  # French (/two or more digits then NOT -M or -S)
+    "ANSS[Ii]-CC-[0-9][0-9][0-9][0-9][/-_][0-9]+[_/-]M[0-9]+",  # French, maintenance report (ANSSI-CC-2014_46_M01)
+    "ANSS[Ii]-CC-[0-9][0-9][0-9][0-9][/-_][0-9]+[_/-]S[0-9]+",  # French, surveillance report (ANSSI-CC-2012/70-S01)
     # 'ANSSI-CC-CER-F-.+?', # French
-    "DCSSI-[0-9]+?/[0-9]+?",  # French
-    "Certification Report [0-9]+?/[0-9]+?",  # French
-    "Rapport de certification [0-9]+?/[0-9]+?",  # French
+    "DCSSI-[0-9]+/[0-9]+",  # French (DCSSI-2009/07)
+    "Certification Report [0-9]+/[0-9]+",  # French or Australia
+    "Rapport de certification [0-9]+/[0-9]+",  # French
     "NSCIB-CC-[0-9][0-9][0-9][0-9].+?",  # Netherlands
     "NSCIB-CC-[0-9][0-9][0-9][0-9][0-9]*-CR",  # Netherlands
     "NSCIB-CC-[0-9][0-9]-[0-9]+?-CR[0-9]+?",  # Netherlands
-    "SERTIT-[0-9]+?",  # Norway
-    "CCEVS-VR-(?:|VID)[0-9]+?-[0-9]+?",  # US NSA
+    "NSCIB-CC-[0-9][0-9]-[0-9]+(-CR[0-9]+)*",  # Netherlands (old number NSCIB-CC-05-6609 or NSCIB-CC-05-6609-CR)
+    "NSCIB-CC-[0-9]+-CR[0-9]*",  # Netherlands (new number NSCIB-CC-111441-CR NSCIB-CC-111441-CR1)
+    "NSCIB-CC-[0-9]+-MA[0-9]*",  # Netherlands (new number NSCIB-CC-222073-MA NSCIB-CC-200716-MA2)
+    "NSCIB-CC-[0-9][0-9]-[0-9]+",  # Netherlands   (old number NSCIB-CC-05-6609)
+    "NSCIB-CC-[0-9][0-9]-[0-9]+-CR[0-9]+",  # Netherlands (NSCIB-CC-year2digits-number-CR)
+    "SERTIT-[0-9]+",  # Norway
+    "CCEVS-VR-(?:|VID)[0-9]+-[0-9]+[a-z]?",  # US NSA (CCEVS-VR-10884-2018 CCEVS-VR-VID10877-2018)
     # '[0-9][0-9\-]+?-CR', # Canada
+    "CRP[0-9]+",  # UK CESG
     "CRP[0-9][0-9][0-9][0-9]*?",  # UK CESG
-    "CERTIFICATION REPORT No. P[0-9]+?",  # UK CESG
-    "20[0-9][0-9]-[0-9]+-INF-[0-9]+?",  # Spain
-    "KECS-CR-[0-9]+?-[0-9]+?",  # Korea
+    "CERTIFICATION REPORT No. P[0-9]+",  # UK CESG
+    "20[0-9][0-9]-[0-9]+-INF-[0-9]+",  # Spain
+    "20[0-9][0-9]-[0-9]+-INF-[0-9]+(.(?:V|v)[0-9]+)*",  # Spain (2006-4-INF-98 v2 or (2006-4-INF-98-v2))
+    "KECS-CR-[0-9]+-[0-9]+",  # Korea KECS-CR-20-61
     "KECS-ISIS-[0-9]+?-[0-9][0-9][0-9][0-9]",  # Korea
+    "KECS-(?:ISIS|NISS|CISS)-[0-9]+-[0-9][0-9][0-9][0-9]",  # Korea
     "CRP-C[0-9]+?-[0-9]+?",  # Japan
+    "(?:CRP|ACR)-C[0-9]+-[0-9]+",  # Japan (CRP-C0595-01 ACR-C0417-03)
+    "JISEC-CC-CRP-C[0-9]+-[0-9]+-[0-9]+",  # Japan (JISEC-CC-CRP-C0689-01-2020)
     "ISCB-[0-9]+?-RPT-[0-9]+?",  # Malaysia
+    "ISCB-[0-9]+-(?:RPT|FRM)-[CM][0-9]+[A-Z]?-(?:CR|AMR)-[vV][0-9][a-z]?",  # Malaysia (ISCB-3-RPT-C092-CR-v1)
     "OCSI/CERT/.+?",  # Italia
-    "[0-9\\.]+?/TSE-CCCS-[0-9]+?",  # Turkis CCCS
-    "BTBD-.+?",  # Turkis CCCS
+    "OCSI/CERT/.+?/20[0-9]+(?:/w|/RC)",  # Italia  (OCSI/CERT/ATS/01/2018/RC)
+    "[0-9\\.]+?/TSE-CCCS-[0-9]+",  # Turkis CCCS (21.0.0sc/TSE-CCCS-75)
+    "CSEC[0-9]+",  # Sweden (CSEC2019015)
+    # India (IC3S/DEL01/VALIANT/EAL1/0317/0007/CR  STQC/CC/14-15/12/ETR/0017 IC3S/MUM01/CISCO/cPP/0119/0016/CR)
+    # will miss STQC/CC/14-15/12/ETR/0017
+    "(?:IC3S|STQC/CC)/[^ ]+?/CR ",  # must end with CR, no space inside
+    "CSA_CC_[0-9]+",  # Singapure (CSA_CC_19001)
+    "[0-9][0-9][0-9]-[47]-[0-9][0-9][0-9](-CR)*",  # Canada xxx-{47}-xxx (383-4-438, 383-4-82-CR)
 ]
 
 rules_vendor = [
