@@ -3,20 +3,21 @@ import html
 import logging
 import os
 import re
-import subprocess
 import time
 from datetime import date, datetime
 from enum import Enum
 from multiprocessing.pool import ThreadPool
 from pathlib import Path
-from typing import Dict, Hashable, List, Optional, Sequence, Set, Tuple, Union
+from typing import Any, Dict, Generator, Hashable, Iterator, List, Optional, Sequence, Set, Tuple, Union
 
 import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
+import pdftotext
 import pikepdf
 import requests
 from PyPDF2 import PdfFileReader
+from PyPDF2.generic import BooleanObject, FloatObject, IndirectObject, NumberObject
 from tqdm import tqdm as tqdm_original
 
 import sec_certs.constants as constants
@@ -34,6 +35,7 @@ from sec_certs.constants import (
 logger = logging.getLogger(__name__)
 
 
+# TODO: Once typehints in tqdm are implemented, we should use them: https://github.com/tqdm/tqdm/issues/260
 def tqdm(*args, **kwargs):
     if "disable" in kwargs:
         return tqdm_original(*args, **kwargs)
@@ -79,9 +81,9 @@ def get_first_16_bytes_sha256(string: str) -> str:
     return hashlib.sha256(string.encode("utf-8")).hexdigest()[:16]
 
 
-def get_sha256_filepath(filepath):
+def get_sha256_filepath(filepath: Union[str, Path]) -> str:
     hash_sha256 = hashlib.sha256()
-    with open(filepath, "rb") as f:
+    with Path(filepath).open("rb") as f:
         for chunk in iter(lambda: f.read(4096), b""):
             hash_sha256.update(chunk)
     return hash_sha256.hexdigest()
@@ -98,12 +100,13 @@ def sanitize_date(record: Union[pd.Timestamp, date, np.datetime64]) -> Union[dat
         return None
     elif isinstance(record, pd.Timestamp):
         return record.date()
-    else:
-        return record  # type: ignore
+    elif isinstance(record, (date, type(None))):
+        return record
+    raise ValueError("Unsupported type given as input")
 
 
 def sanitize_string(record: str) -> str:
-    # TODO: There is a sample with name 'ATMEL Secure Microcontroller AT90SC12872RCFT &#x2f; AT90SC12836RCFT rev. I &amp;&#x23;38&#x3b; J' that has to be unescaped twice
+    # There is a sample with name 'ATMEL Secure Microcontroller AT90SC12872RCFT &#x2f; AT90SC12836RCFT rev. I &amp;&#x23;38&#x3b; J' that has to be unescaped twice
     string = html.unescape(html.unescape(record)).replace("\n", "")
     return " ".join(string.split())
 
@@ -127,7 +130,6 @@ def sanitize_protection_profiles(record: str) -> list:
     return record.split(",")
 
 
-# TODO: realize whether this stays or goes somewhere else
 def parse_list_of_tables(txt: str) -> Set[str]:
     """
     Parses list of tables from function find_tables(), finds ones that mention algorithms
@@ -161,7 +163,7 @@ def find_tables_iterative(file_text: str) -> List[int]:
     return list(pages)
 
 
-def find_tables(txt: str, file_name: Path) -> Optional[List]:
+def find_tables(txt: str, file_name: Path) -> Optional[Union[List[str], List[int]]]:
     """
     Function that tries to pages in security policy pdf files, where it's possible to find a table containing
     algorithms
@@ -185,7 +187,7 @@ def find_tables(txt: str, file_name: Path) -> Optional[List]:
     return table_page_indices if table_page_indices else None
 
 
-def repair_pdf(file: Path):
+def repair_pdf(file: Path) -> None:
     """
     Some pdfs can't be opened by PyPDF2 - opening them with pikepdf and then saving them fixes this issue.
     By opening this file in a pdf reader, we can already extract number of pages
@@ -196,32 +198,59 @@ def repair_pdf(file: Path):
     pdf.save(file)
 
 
-def convert_pdf_file(pdf_path: Path, txt_path: Path, options):
-    response = subprocess.run(
-        ["pdftotext", *options, pdf_path, txt_path], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, timeout=60
-    ).returncode
-    if response == 0:
-        return constants.RETURNCODE_OK
-    else:
+def convert_pdf_file(pdf_path: Path, txt_path: Path) -> str:
+    try:
+        with pdf_path.open("rb") as pdf_handle:
+            pdf = pdftotext.PDF(pdf_handle, "", True)  # No password, Raw=True
+            txt = "".join(pdf)
+    except Exception as e:
+        logger.error(f"Error when converting pdf->txt: {e}")
         return constants.RETURNCODE_NOK
 
+    with txt_path.open("w", encoding="utf-8") as txt_handle:
+        txt_handle.write(txt)
+
+    return constants.RETURNCODE_OK
+
+
+def extract_pdf_metadata(filepath: Path) -> Tuple[str, Optional[Dict[str, Any]]]:
+    def map_metadata_value(val, nope_out=False):
+        if isinstance(val, BooleanObject):
+            val = val.value
+        elif isinstance(val, FloatObject):
+            val = float(val)
+        elif isinstance(val, NumberObject):
+            val = int(val)
+        elif isinstance(val, IndirectObject) and not nope_out:
+            # Let's make sure to nope out in case of cycles
+            val = map_metadata_value(val.getObject(), nope_out=True)
+        else:
+            val = str(val)
+        return val
 
-def extract_pdf_metadata(filepath: Path):
     metadata = dict()
 
     try:
         metadata["pdf_file_size_bytes"] = filepath.stat().st_size
         with filepath.open("rb") as handle:
-            pdf = PdfFileReader(handle)
-
+            pdf = PdfFileReader(handle, strict=False)
             metadata["pdf_is_encrypted"] = pdf.getIsEncrypted()
-            metadata["pdf_number_of_pages"] = pdf.getNumPages()
 
+        # see https://stackoverflow.com/questions/26242952/pypdf-2-decrypt-not-working
+        if metadata["pdf_is_encrypted"]:
+            pikepdf.open(filepath, allow_overwriting_input=True).save()
+
+        with filepath.open("rb") as handle:
+            pdf = PdfFileReader(handle, strict=False)
+            metadata["pdf_number_of_pages"] = pdf.getNumPages()
             pdf_document_info = pdf.getDocumentInfo()
-            metadata.update({key: str(val) for key, val in pdf_document_info.items()} if pdf_document_info else {})
+
+        for key, val in pdf_document_info.items():
+            metadata[str(key)] = map_metadata_value(val)
 
     except Exception as e:
-        error_msg = f"Failed to read metadata of {filepath}, error: {e}"
+        relative_filepath = "/".join(str(filepath).split("/")[-4:])
+        error_msg = f"Failed to read metadata of {relative_filepath}, error: {e}"
         logger.error(error_msg)
         return error_msg, None
 
@@ -484,7 +513,8 @@ def search_only_headers_anssi(filepath: Path):  # noqa: C901
                 items_found[constants.TAG_CERT_LAB] = normalize_match_string(match_groups[index_next_item])
                 index_next_item += 1
     except Exception as e:
-        error_msg = f"Failed to parse ANSSI frontpage headers from {filepath}; {e}"
+        relative_filepath = "/".join(str(filepath).split("/")[-4:])
+        error_msg = f"Failed to parse ANSSI frontpage headers from {relative_filepath}; {e}"
         logger.error(error_msg)
         return error_msg, None
 
@@ -592,7 +622,212 @@ def search_only_headers_bsi(filepath: Path):  # noqa: C901
         # print('Total no hits files: {}'.format(len(files_without_match)))
         # print('\n**********************************')
     except Exception as e:
-        error_msg = f"Failed to parse BSI headers from frontpage: {filepath}; {e}"
+        relative_filepath = "/".join(str(filepath).split("/")[-4:])
+        error_msg = f"Failed to parse BSI headers from frontpage: {relative_filepath}; {e}"
+        logger.error(error_msg)
+        return error_msg, None
+
+    return constants.RETURNCODE_OK, items_found
+
+
+# Port from old-api branch
+def search_only_headers_nscib(filepath: Path):  # noqa: C901
+    LINE_SEPARATOR_STRICT = " "
+    NUM_LINES_TO_INVESTIGATE = 60
+    items_found: Dict[str, str] = {}
+
+    try:
+        # Process front page with info: cert_id, certified_item and developer
+        whole_text, whole_text_with_newlines, was_unicode_decode_error = load_cert_file(
+            filepath, NUM_LINES_TO_INVESTIGATE, LINE_SEPARATOR_STRICT
+        )
+
+        certified_item = ""
+        developer = ""
+        cert_lab = ""
+        cert_id = ""
+
+        lines = whole_text_with_newlines.splitlines()
+        no_match_yet = True
+        item_offset = -1
+
+        for line_index in range(0, len(lines)):
+            line = lines[line_index]
+
+            if "Certification Report" in line:
+                item_offset = line_index + 1
+            if "Assurance Continuity Maintenance Report" in line:
+                item_offset = line_index + 1
+
+            SPONSORDEVELOPER_STR = "Sponsor and developer:"
+
+            if SPONSORDEVELOPER_STR in line:
+                if no_match_yet:
+                    items_found = {}
+                    no_match_yet = False
+
+                # all lines above till 'Certification Report' or 'Assurance Continuity Maintenance Report'
+                certified_item = ""
+                for name_index in range(item_offset, line_index):
+                    certified_item += lines[name_index] + " "
+                developer = line[line.find(SPONSORDEVELOPER_STR) + len(SPONSORDEVELOPER_STR) :]
+
+            SPONSOR_STR = "Sponsor:"
+
+            if SPONSOR_STR in line:
+                if no_match_yet:
+                    items_found = {}
+                    no_match_yet = False
+
+                # all lines above till 'Certification Report' or 'Assurance Continuity Maintenance Report'
+                certified_item = ""
+                for name_index in range(item_offset, line_index):
+                    certified_item += lines[name_index] + " "
+
+            DEVELOPER_STR = "Developer:"
+            if DEVELOPER_STR in line:
+                developer = line[line.find(DEVELOPER_STR) + len(DEVELOPER_STR) :]
+
+            CERTLAB_STR = "Evaluation facility:"
+            if CERTLAB_STR in line:
+                cert_lab = line[line.find(CERTLAB_STR) + len(CERTLAB_STR) :]
+
+            REPORTNUM_STR = "Report number:"
+            if REPORTNUM_STR in line:
+                cert_id = line[line.find(REPORTNUM_STR) + len(REPORTNUM_STR) :]
+
+        if not no_match_yet:
+            items_found[constants.TAG_CERT_ID] = normalize_match_string(cert_id)
+            items_found[constants.TAG_CERT_ITEM] = normalize_match_string(certified_item)
+            items_found[constants.TAG_DEVELOPER] = normalize_match_string(developer)
+            items_found[constants.TAG_CERT_LAB] = cert_lab
+
+    except Exception as e:
+        error_msg = f"Failed to parse NSCIB headers from frontpage: {filepath}; {e}"
+        logger.error(error_msg)
+        return error_msg, None
+
+    return constants.RETURNCODE_OK, items_found
+
+
+# Port from old-api branch
+def search_only_headers_niap(filepath: Path):
+    LINE_SEPARATOR_STRICT = " "
+    NUM_LINES_TO_INVESTIGATE = 15
+    items_found: Dict[str, str] = {}
+
+    try:
+        # Process front page with info: cert_id, certified_item and developer
+        whole_text, whole_text_with_newlines, was_unicode_decode_error = load_cert_file(
+            filepath, NUM_LINES_TO_INVESTIGATE, LINE_SEPARATOR_STRICT
+        )
+
+        certified_item = ""
+        cert_id = ""
+
+        lines = whole_text_with_newlines.splitlines()
+        no_match_yet = True
+        item_offset = -1
+
+        for line_index in range(0, len(lines)):
+            line = lines[line_index]
+
+            if "Validation Report" in line:
+                item_offset = line_index + 1
+
+            REPORTNUM_STR = "Report Number:"
+            if REPORTNUM_STR in line:
+                if no_match_yet:
+                    items_found = {}
+                    no_match_yet = False
+
+                # all lines above till 'Certification Report' or 'Assurance Continuity Maintenance Report'
+                certified_item = ""
+                for name_index in range(item_offset, line_index):
+                    certified_item += lines[name_index] + " "
+                cert_id = line[line.find(REPORTNUM_STR) + len(REPORTNUM_STR) :]
+                break
+
+        if not no_match_yet:
+            items_found[constants.TAG_CERT_ID] = normalize_match_string(cert_id)
+            items_found[constants.TAG_CERT_ITEM] = normalize_match_string(certified_item)
+            items_found[constants.TAG_CERT_LAB] = "US NIAP"
+
+    except Exception as e:
+        error_msg = f"Failed to parse NIAP headers from frontpage: {filepath}; {e}"
+        logger.error(error_msg)
+        return error_msg, None
+
+    return constants.RETURNCODE_OK, items_found
+
+
+# Port from old-api branch
+def search_only_headers_canada(filepath: Path):  # noqa: C901
+    LINE_SEPARATOR_STRICT = " "
+    NUM_LINES_TO_INVESTIGATE = 20
+    items_found: Dict[str, str] = {}
+    try:
+        whole_text, whole_text_with_newlines, was_unicode_decode_error = load_cert_file(
+            filepath, NUM_LINES_TO_INVESTIGATE, LINE_SEPARATOR_STRICT
+        )
+
+        cert_id = ""
+
+        lines = whole_text_with_newlines.splitlines()
+        no_match_yet = True
+        for line_index in range(0, len(lines)):
+            line = lines[line_index]
+            if "Government of Canada, Communications Security Establishment" in line:
+                REPORTNUM_STR1 = "Evaluation number:"
+                REPORTNUM_STR2 = "Document number:"
+                matched_number_str = ""
+                line_certid = lines[line_index + 1]
+                if line_certid.startswith(REPORTNUM_STR1):
+                    matched_number_str = REPORTNUM_STR1
+                if line_certid.startswith(REPORTNUM_STR2):
+                    matched_number_str = REPORTNUM_STR2
+                if matched_number_str != "":
+                    if no_match_yet:
+                        items_found = {}
+                        no_match_yet = False
+
+                    cert_id = line_certid[line_certid.find(matched_number_str) + len(matched_number_str) :]
+                    break
+
+            if (
+                "Government of Canada. This document is the property of the Government of Canada. It shall not be altered,"
+                in line
+            ):
+                REPORTNUM_STR = "Evaluation number:"
+                for offset in range(1, 20):
+                    line_certid = lines[line_index + offset]
+                    if "UNCLASSIFIED" in line_certid:
+                        if no_match_yet:
+                            items_found = {}
+                            no_match_yet = False
+                        line_certid = lines[line_index + offset - 4]
+                        cert_id = line_certid[line_certid.find(REPORTNUM_STR) + len(REPORTNUM_STR) :]
+                        break
+                if not no_match_yet:
+                    break
+
+            if (
+                "UNCLASSIFIED / NON CLASSIFIÉ" in line
+                and "COMMON CRITERIA CERTIFICATION REPORT" in lines[line_index + 2]
+            ):
+                line_certid = lines[line_index + 1]
+                if no_match_yet:
+                    items_found = {}
+                    no_match_yet = False
+                cert_id = line_certid
+                break
+
+        if not no_match_yet:
+            items_found[constants.TAG_CERT_ID] = normalize_match_string(cert_id)
+            items_found[constants.TAG_CERT_LAB] = "CANADA"
+
+    except Exception as e:
+        error_msg = f"Failed to parse Canada headers from frontpage: {filepath}; {e}"
         logger.error(error_msg)
         return error_msg, None
 
@@ -609,7 +844,8 @@ def extract_keywords(filepath: Path) -> Tuple[str, Optional[Dict[str, Dict[str,
             processed_result[key] = {key: val for key, val in gen_dict_extract(result[key])}
 
     except Exception as e:
-        error_msg = f"Failed to parse keywords from: {filepath}; {e}"
+        relative_filepath = "/".join(str(filepath).split("/")[-4:])
+        error_msg = f"Failed to parse keywords from: {relative_filepath}; {e}"
         logger.error(error_msg)
         return error_msg, None
     return constants.RETURNCODE_OK, processed_result
@@ -624,7 +860,7 @@ def plot_dataframe_graph(
     bins: int = 50,
     log: bool = True,
     show: bool = True,
-):
+) -> None:
     pd_data = pd.Series(data)
     pd_data.hist(bins=bins, label=label, density=density, cumulative=cumulative)
     plt.savefig(file_name)
@@ -637,7 +873,7 @@ def plot_dataframe_graph(
     logger.info(sorted_data.where(sorted_data > 1).dropna())
 
 
-def is_in_dict(target_dict, path):
+def is_in_dict(target_dict: Dict, path: str) -> bool:
     current_level = target_dict
     for item in path:
         if item not in current_level:
@@ -647,16 +883,16 @@ def is_in_dict(target_dict, path):
     return True
 
 
-def search_files(folder):
-    for root, dirs, files in os.walk(folder):
+def search_files(folder: str) -> Iterator[str]:
+    for root, _, files in os.walk(folder):
         yield from [os.path.join(root, x) for x in files]
 
 
-def save_modified_cert_file(target_file, modified_cert_file_text, is_unicode_text):
+def save_modified_cert_file(target_file: Union[str, Path], modified_cert_file_text: str, is_unicode_text: bool) -> None:
     if is_unicode_text:
-        write_file = open(target_file, "w", encoding="utf8", errors="replace")
+        write_file = Path(target_file).open("w", encoding="utf8", errors="replace")
     else:
-        write_file = open(target_file, "w", errors="replace")
+        write_file = Path(target_file).open("w", errors="replace")
 
     try:
         write_file.write(modified_cert_file_text)
@@ -738,15 +974,17 @@ def parse_cert_file(file_name, search_rules, limit_max_lines=-1, line_separator=
     return items_found_all, (whole_text_with_newlines, was_unicode_decode_error)
 
 
-def normalize_match_string(match):
+def normalize_match_string(match: str) -> str:
     match = match.strip().rstrip('];.”":)(,').rstrip(os.sep).replace("  ", " ")
     return "".join(filter(str.isprintable, match))
 
 
-def load_cert_file(file_name, limit_max_lines=-1, line_separator=LINE_SEPARATOR):
+def load_cert_file(
+    file_name: Union[str, Path], limit_max_lines: int = -1, line_separator: str = LINE_SEPARATOR
+) -> Tuple[str, str, bool]:
     lines = []
     was_unicode_decode_error = False
-    with open(file_name, "r", errors=FILE_ERRORS_STRATEGY) as f:
+    with Path(file_name).open("r", errors=FILE_ERRORS_STRATEGY) as f:
         try:
             lines = f.readlines()
         except UnicodeDecodeError:
@@ -785,7 +1023,7 @@ def load_cert_file(file_name, limit_max_lines=-1, line_separator=LINE_SEPARATOR)
     return whole_text, whole_text_with_newlines, was_unicode_decode_error
 
 
-def load_cert_html_file(file_name):
+def load_cert_html_file(file_name: str) -> str:
     with open(file_name, "r", errors=FILE_ERRORS_STRATEGY) as f:
         try:
             whole_text = f.read()
@@ -799,7 +1037,7 @@ def load_cert_html_file(file_name):
     return whole_text
 
 
-def gen_dict_extract(dct: Dict, searched_key: Hashable = "count"):
+def gen_dict_extract(dct: Dict, searched_key: Hashable = "count") -> Generator[Any, None, None]:
     """
     Function to flatten dictionary with some serious limitations. We only expect to use it temporarily on dictionary
     produced by extract_keywords that contains many layers. On the deepest level in that dictionary, 'some_match': {'count': frequency}.
@@ -819,7 +1057,7 @@ def gen_dict_extract(dct: Dict, searched_key: Hashable = "count"):
                     yield key, result
 
 
-def compute_heuristics_version(cert_name: str) -> List[str]:
+def compute_heuristics_version(cert_name: str) -> Set[str]:
     """
     Will extract possible versions from the name of sample
     """
@@ -844,10 +1082,10 @@ def compute_heuristics_version(cert_name: str) -> List[str]:
     # return identified_versions if identified_versions else ['-']
 
     if not matches:
-        return ["-"]
+        return {constants.CPE_VERSION_NA}
 
     matched = [re.search(normalizer, x) for x in matches]
-    return [x.group() for x in matched if x is not None]
+    return {x.group() for x in matched if x is not None}
 
 
 def tokenize_dataset(dset: List[str], keywords: Set[str]) -> np.ndarray:
@@ -858,7 +1096,38 @@ def tokenize(string: str, keywords: Set[str]) -> str:
     return " ".join([x for x in string.split() if x.lower() in keywords])
 
 
-def filter_shortly_described_cves(x, y):
-    n_tokens = np.array(list(map(lambda item: len(item.split(" ")), x)))
-    indices = np.where(n_tokens > 5)
-    return np.array(x)[indices], np.array(y)[indices]
+# Credit: https://stackoverflow.com/questions/18092354/
+def split_unescape(s: str, delim: str, escape: str = "\\", unescape: bool = True) -> List[str]:
+    """
+    >>> split_unescape('foo,bar', ',')
+    ['foo', 'bar']
+    >>> split_unescape('foo$,bar', ',', '$')
+    ['foo,bar']
+    >>> split_unescape('foo$$,bar', ',', '$', unescape=True)
+    ['foo$', 'bar']
+    >>> split_unescape('foo$$,bar', ',', '$', unescape=False)
+    ['foo$$', 'bar']
+    >>> split_unescape('foo$', ',', '$', unescape=True)
+    ['foo$']
+    """
+    ret = []
+    current = []
+    itr = iter(s)
+    for ch in itr:
+        if ch == escape:
+            try:
+                # skip the next character; it has been escaped!
+                if not unescape:
+                    current.append(escape)
+                current.append(next(itr))
+            except StopIteration:
+                if unescape:
+                    current.append(escape)
+        elif ch == delim:
+            # split! (add current to the list and reset it)
+            ret.append("".join(current))
+            current = []
+        else:
+            current.append(ch)
+    ret.append("".join(current))
+    return ret
